import gc
from typing import List

import numpy as np

import matplotlib
import matplotlib.pyplot as plt

from .models import Episode
from .serializer import episodes_to_df

def create_heatmap(episodes: list[Episode]) -> np.ndarray:
    # read ratings for series
    df = episodes_to_df(episodes)

    # extract dimension of series
    seasons = df['season'].max()
    episodes = df['episode'].max()

    # initialize heatmap with zeros
    heatmap = np.zeros([seasons, episodes])

    for _, row in df.iterrows():
        # indexing in df starts at 1, convert to 0-indexing
        sid = int(row['season'])-1
        eid = int(row['episode'])-1
        rating = row['rating']

        # write rating in heatmap
        heatmap[seasons - sid - 1, eid] = rating

    # maks out unused squares
    heatmap = np.ma.masked_where(heatmap == 0.0, heatmap)

    return heatmap

def heatmap_plot(series_name: str, episodes: List[Episode], show: bool, save_fn: str or None,
                 dark_mode: bool, save_buffer=None, save_buffer_format='png', dpi=100, dark_col=(0.16, 0.16, 0.16, 1.0)) -> None:
    # create the heatmap array
    heatmap = create_heatmap(episodes)

    if show:
        matplotlib.use('TkAgg')

    # scale figure appropriately
    figsize = np.array(heatmap.shape[::-1])
    figsize = figsize/3
    minax = np.min(figsize)
    minsize = 2.3
    if minax < minsize:
        figsize = figsize / minax * minsize
    fig = plt.figure(figsize=figsize, dpi=dpi)

    # dark mode
    if dark_mode:
        plt.style.use("dark_background")
        plt.gca().set_facecolor(dark_col)
        plt.gcf().set_facecolor(dark_col)

    # plot heatmap and inner ratings text
    plt.imshow(heatmap)
    for sid in range(heatmap.shape[0]):
        for eid in range(heatmap.shape[1]):
            rating = heatmap[sid, eid]
            plt.text(eid, sid, f'{rating}', ha='center', va='center', color='w', fontsize=7)

    # format tick labels
    ticks = [np.arange(heatmap.shape[i]) for i in [0, 1]]
    plt.yticks(ticks[0][::-1], labels=[f'{i+1}' for i in ticks[0]])
    plt.xticks(ticks[1], labels=[f'{i+1}' for i in ticks[1]])

    # format axis labels
    plt.xlabel('Episode')
    plt.ylabel('Season')

    # title with whitespaces
    title = series_name.replace('_', ' ')
    plt.title(title)

    # plot heatmap
    plt.tight_layout()

    # save plot to buffer if provided
    if save_buffer is not None:
        plt.savefig(save_buffer, format=save_buffer_format, dpi=dpi)
        save_buffer.seek(0)

    # optionally save or display the plot
    if save_fn is not None:
        if dark_mode:
            plt.savefig(save_fn, facecolor=dark_col, dpi=dpi)
        else:
            plt.savefig(save_fn, dpi=dpi)
    if show:
        plt.show()

    fig.clf()
    plt.close()
    gc.collect()



from typing import List

from lxml import html

from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By

from .models import Episode, ImdbEntry

class ImdbScraper:
    timeout = 10
    max_retries = 20
    max_episodes = 100
    base_url = 'https://www.imdb.com'
    driver = None
    variant = None

    xpaths = {
        'result_list': [
            '''//*[@id='__next']/main/div[2]/div[3]/section/div/div[1]/section[3]/div[2]/ul''',
            '//*[@id="main"]/div/div[2]/table/tbody'
        ],
        'entry_id': [
            '//*[@id="__next"]/main/div[2]/div[3]/section/div/div[1]/section[2]/div[2]/ul/li[{}]/div[2]/div[1]/a',
            '//*[@id="main"]/div/div[2]/table/tbody/tr[{}]/td[2]/a'
        ],
        'entry_year': [
            '//*[@id="__next"]/main/div[2]/div[3]/section/div/div[1]/section[2]/div[2]/ul/li[{}]/div[2]/div[1]/ul[1]/li[1]/label',
            '//*[@id="main"]/div/div[2]/table/tbody/tr[{}]/td[2]'
        ],
        'num_seasons': [
            '''//*[@id='bySeason']'''
        ]
    }

    def __init__(self):
        self.chrome_options = webdriver.ChromeOptions()
        self.chrome_options.add_argument('--headless')
        self.chrome_options.add_argument('window-size=1920x1080')

    def __enter__(self):
        self.driver = webdriver.Chrome(options=self.chrome_options)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.driver.close()

    def _restart_driver(self) -> None:
        self.page_loaded = False
        self.driver.delete_all_cookies()
        self.driver.close()
        self.driver = webdriver.Chrome(options=self.chrome_options)

    def _load_url(self, imdb_url: str) -> None:
        self.variant = None
        print(f'Fetching from {imdb_url}')
        #self.driver.implicitly_wait(1)
        self.driver.get(f'{self.base_url}/{imdb_url}')

    def _load_element(self, xpaths):
        if self.variant is None:
            try:
                element_present = EC.presence_of_element_located((By.XPATH, xpaths[0]))
                WebDriverWait(self.driver, timeout=self.timeout).until(element_present)
                self.variant = 0
            except TimeoutException:
                self.variant = min(1, len(xpaths)-1)

            print(f'Loaded variant {self.variant}')

        # prefer to use variant indexed xpath, but check others if no match
        xp_var = xpaths[self.variant]
        prioritized_xpaths = [xp_var, *[xp for xp in xpaths if xp != xp_var]]

        for xp in prioritized_xpaths:
            tree = html.fromstring(self.driver.page_source)
            matches = tree.xpath(xp)
            if len(matches) == 0:
                continue
            self.variant = xpaths.index(xp)
            return matches[0]

        return None

    def _find_element_initial(self, imdb_url: str, xpaths):
        for attempt in range(self.max_retries):
            print(f'find_element attempt ({attempt+1}/{self.max_retries})')
            self._load_url(imdb_url)
            elem = self._load_element(xpaths)
            if elem is None:
                self._restart_driver()
                continue
            return elem
        print('Failed to find xpath element.')
        return None

    def _xpf(self, xpid, arg) -> List[str]:
        return [p.format(arg) for p in self.xpaths[xpid]]

    def search(self, search_text: str, search_type: str = None, max_results: int = 5) -> List[ImdbEntry]:
        results = []

        # [None, 'tv', 'ft']
        type_query = ''
        if search_type is not None:
            type_query = f'&s=tt&type={search_type}'

        search_text = search_text.replace(' ', '%20')

        search_results = self._find_element_initial(
            imdb_url=f'/find?q={search_text}{type_query}',
            xpaths=self.xpaths['result_list']
        )

        # //*[@id="__next"]/main/div[2]/div[3]/section/div/div[1]/section[2]/div[2]/ul --- if matches, ul children <= 5
        num_results = len(search_results)

        # cap results at max_results
        num_results = min(max_results, num_results)

        # individual results
        # //*[@id="__next"]/main/div[2]/div[3]/section/div/div[1]/section[2]/div[2]/ul/li[1]/div[2]/div[1]/a
        # //*[@id="__next"]/main/div[2]/div[3]/section/div/div[1]/section[2]/div[2]/ul/li[2]/div[2]/div[1]/a
        # xpath=f'//*[@id="__next"]/main/div[2]/div[3]/section/div/div[1]/section[2]/div[2]/ul/li[{li_index}]/div[2]/div[1]/a'

        for i in range(num_results):
            li_index = i+1

            # extract href from <a> tag
            id_elem = self._load_element(
                xpaths=self._xpf('entry_id', li_index)
            )
            entry_name = id_elem.text

            entry_id = id_elem.attrib['href'].split('/')[2]

            year_elem = self._load_element(
                xpaths=self._xpf('entry_year', li_index)
            )

            if self.variant == 0:
                year_text = year_elem.text
                if year_elem is None:
                    entry_year = 0
                    entry_type = 'na'
                else:
                    year_splits = year_text.replace(' ', '').split('–')
                    entry_type = 'ft' if len(year_splits) == 1 else 'tv'
                    entry_year = int(year_splits[0])
            else:
                info_text = year_elem[0].tail.replace(' ', '')
                infos = info_text.split('(')
                if len(infos) == 1:
                    entry_year = 0
                    entry_type = 'na'
                else:
                    entry_year = infos[1][:infos[1].index(')')]
                    if len(infos) == 2:
                        entry_type = 'ft'
                    else:
                        entry_type = infos[2][:infos[2].index(')')]
                        entry_type = 'tv' if entry_type == 'TVSeries' else 'ft'

            if search_type is not None and search_type != entry_type:
                print(f'WARNING: found an entry of type {entry_type} but search query was for type {search_type}')

            results.append(ImdbEntry(
                id=entry_id,
                title=entry_name,
                year=entry_year,
                type=entry_type
            ))

        return results

    def get_all_episodes(self, series_id: str) -> List[Episode]:
        episodes = []

        seasons_dropdown = self._find_element_initial(
            imdb_url=f'/title/{series_id}/episodes?season=1',
            xpaths=self.xpaths['num_seasons']
        )

        num_seasons = 0
        for seas_num, val in enumerate(seasons_dropdown.value_options):
            if str(seas_num + 1) == val:
                num_seasons += 1

        print(f'Found {num_seasons} seasons')

        for i in range(num_seasons):
            season_num = i + 1
            episodes.extend(self.get_episodes(series_id, season_num))

        return episodes

    def get_episodes(self, series_id: str, season_num: int) -> List[Episode]:
        episodes = []

        self._find_element_initial(
            imdb_url=f'/title/{series_id}/episodes?season={season_num}',
            xpaths=[f'''//*[@id='episodes_content']/div[2]/div[2]/div[{1}]/div[2]/div[2]/div[1]/span[2]''']
        )

        ep = 0
        try:
            for ep in range(0, self.max_episodes):
                episode_num = ep+1

                rating_elem = self._load_element(
                    xpaths=[f'''//*[@id='episodes_content']/div[2]/div[2]/div[{episode_num}]/div[2]/div[2]/div[1]/span[2]''']
                )
                if rating_elem is None:
                    break
                rating = float(rating_elem.text)

                name_elem = self._load_element(
                    xpaths=[f'//*[@id="episodes_content"]/div[2]/div[2]/div[{episode_num}]/div[2]/strong/a']
                )
                ep_name = name_elem.text

                print(f'S{season_num} E{episode_num} rating={rating:.1f} name="{ep_name}"')
                episodes.append(Episode(
                    season_num=season_num,
                    episode_num=episode_num,
                    name=ep_name,
                    rating=rating,
                ))
        except IndexError:
            pass
        print(f'Found {ep} episodes in season {season_num}')

        return episodes



import argparse
import os

import pandas as pd
from tqdm import tqdm

from seriesheatmap.heatmap import heatmap_plot
from seriesheatmap.serializer import df_to_episodes


def to_fn(series: str) -> str:
    return series.replace(' ', '_')

def fn_match(f1: str, f2: str):
    f1 = f1.lower().replace(' ', '').replace('_', '').replace(':', '')
    f2 = f2.lower().replace(' ', '').replace('_', '').replace(':', '')
    return f1 in f2 or f2 in f1

def main():
    # define and parse args
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--show', action='store_true', help='show the heatmap plot instead of saving it')
    parser.add_argument('-d', '--dark', action='store_true', help='use dark mode for the plot style')
    parser.add_argument('-o', '--override', action='store_true', help='override existing plots')
    parser.add_argument('-n', '--name', type=str, default=None,
                        help='name of the series, if not set the whole data directory will be scanned')
    args = parser.parse_args()

    # config from args
    show = args.show
    save = not show
    override = args.override
    dark_mode = args.dark
    custom_series = args.name

    # paths
    img_path = 'examples/img'
    data_path = 'examples/data'

    # if custom series name supplied from arguments, use this. otherwise load all series from data directory
    all_series = ['.'.join(s.split('.')[:-1]) for s in os.listdir('examples/data')]
    if custom_series is not None:
        series = [custom_series]
    else:
        # load all series in data folder
        series = all_series

    # create heatmap plots
    if len(series) == 0:
        print('No series to plot...')
    else:
        for s in tqdm(series):
            fn_s = to_fn(s)
            fn_data = f'{data_path}/{fn_s}.csv'
            if not os.path.isfile(fn_data):
                # try to find closest match
                matches = [m for m in all_series if fn_match(s, m)]
                if len(matches) == 0:
                    print(f'No match found in data for {s}.')
                    continue
                fn_data = f'{data_path}/{to_fn(matches[0])}.csv'
                fn_s = matches[0]
            df = pd.read_csv(fn_data)
            episodes = df_to_episodes(df)

            # don't override file if already exists and flag is not set
            save_fn = f'{img_path}/{fn_s}.png'
            if os.path.isfile(save_fn) and not override:
                save_fn = None

            heatmap_plot(series_name=s, episodes=episodes, show=show, save_fn=save_fn,
                         dark_mode=dark_mode, dpi=150)

if __name__ == '__main__':
    main()



from typing import List

import pandas as pd

from .models import Episode


def episodes_to_df(episodes: List[Episode]) -> pd.DataFrame:
    df = pd.DataFrame()
    for ep in episodes:
        df = pd.concat([df, pd.DataFrame([{
            'season': ep.season_num,
            'episode': ep.episode_num,
            'name': ep.name,
            'rating': ep.rating
        }])], ignore_index=True)
    return df


def df_to_episodes(df: pd.DataFrame) -> List[Episode]:
    episodes = []
    for _, row in df.iterrows():
        episodes.append(Episode(
            season_num=int(row['season']),
            episode_num=int(row['episode']),
            name=row['name'],
            rating=row['rating'],
        ))
    return episodes



import os
import argparse

from seriesheatmap.scraper import ImdbScraper
from seriesheatmap.serializer import episodes_to_df

def main():
    # define and parse args
    parser = argparse.ArgumentParser()
    parser.add_argument('-n', '--name', type=str, help='name of the series to scrape')
    parser.add_argument('-f', '--file', type=str, help='path to a file with a list of series to scrape')
    args = parser.parse_args()

    # extract query
    queries = []

    if args.name is not None:
        queries.append(args.name)

    if args.file is not None:
        if not os.path.isfile(args.file):
            raise ValueError('Supplied argument file does not exist')

        with open(args.file, 'r') as f:
            for line in f.readlines():
                # ignore empty or commented lines
                l_trimmed = line.replace(' ', '')
                if len(l_trimmed) > 0 and not l_trimmed.startswith('#'):
                    queries.append(line)

    if len(queries) == 0:
        print('No query supplied. Use --help to see how to usae this command.')
    else:
        print(f'Scraping {len(queries)} series')
        for query in queries:
            with ImdbScraper() as sc:
                entries = sc.search(search_text=query, search_type='tv', max_results=5)
                series_entries = [e for e in entries if e.type == 'tv']
                if len(series_entries) == 0:
                    closest_match_text = ''
                    if len(entries) > 0:
                        e = entries[0]
                        closest_match_text = f'Closest match: id="{e.id}" name="{e.title}" type="{e.type}"'
                    print(f'Found no imdb entry for query "{query}". {closest_match_text}')
                    continue
                imdb_entry = series_entries[0]
                title = imdb_entry.title
                episodes = sc.get_all_episodes(imdb_entry.id)

            df = episodes_to_df(episodes)

            fn = title.replace(' ', '_').replace(':', '')
            df.to_csv(f'examples/data/{fn}.csv', index=False)

if __name__ == '__main__':
    main()



from dataclasses import dataclass


@dataclass
class ImdbEntry:
    id: str
    title: str
    year: int
    type: str

@dataclass
class Episode:
    name: str
    season_num: int
    episode_num: int
    rating: float
